-- Forlorn Fox
-- Copyright © 2017 Rob Probin.
-- No redistribution without permission.
--

--local check_debugger = require("debug_select")
--check_debugger("boot")
local exec_functions = require("utilities/exec_functions")

function gulp.boot(app)
    local STARTUP_PAUSE_FOR_DEBUGGER = false
    if STARTUP_PAUSE_FOR_DEBUGGER then
        usleep(5000 * 1000)
    end

    local argc = #arg
    
    local renderer_index = -1;    -- index of rendering driver to use, or -1 for the first one supporting the requested flags
    if argc == 1 and arg[1] == "-v" then
        
        -- print version
        print(string.format("%s", gulp.forlorn_fox_engine_version))
        return 0

    elseif argc >= 2 then

        for i = 1, (argc-1) do  -- (argc-1) because needs parameter afterwards
            if arg[i] == "--renderer" then
                renderer_index = tonumber(arg[i+1])
                -- -1 to initialize the first one supporting the requested flags
                if not renderer_index then renderer_index = -1 end
                print(string.format("Selecting renderer: %d", renderer_index))
            end
        end
    end

    --print(gulp_cpp.SaveDataPath:get_base_pref_path())
    
    -- allow override of switching
    local k = gulp_cpp.SDL_GetModState()
    if (bit32.band(gulp.SDL_Keymod.KMOD_LSHIFT, k) or bit32.band(gulp.SDL_Keymod.KMOD_RSHIFT, k)) == 0 then
        exec_functions.execute_run_alt()
    end
    
    --print(string.format("%.4f ms\n", gulp_cpp.get_time_since_last_call()))
    --gulp_cpp.print_time_since_last_call()
    
    -- SDL_INIT_EVENTS is initialised with a call to SDL_INIT_VIDEO
    local s = gulp.SDL_SubSystems
    if gulp_cpp.SDL_InitSubSystem(s.SDL_INIT_VIDEO + s.SDL_INIT_AUDIO + s.SDL_INIT_TIMER) ~= 0 then
        gulp_cpp.debugMessage("Could initialize extra SDL services\n")
        return 1
    end

    -- seed random number generator
    gulp_cpp.initialise_random_seed()


    -- create window and renderer
    if (not gulp_cpp.CreateGameWindow()) then
        gulp_cpp.debugMessage("Could not create game window\n")
        return 1
    end
    
    local vsync_guess = true
    local r = gulp.SDL_RendererFlags
    local renderer = gulp_cpp.SDL_CreateRenderer(gulp_cpp.GetGameWindow(), renderer_index, bit32.bor(r.SDL_RENDERER_PRESENTVSYNC, r.SDL_RENDERER_ACCELERATED))
    if not renderer then

        gulp_cpp.debugMessage("Accelerated+Vsync renderer not avaiable. SDL says:\n")
        gulp_cpp.debugMessage(SDL_GetError())
        gulp_cpp.debugMessage("\n")
        --accept vsync renderer
        renderer = gulp_cpp.SDL_CreateRenderer(GetGameWindow(), renderer_index, r.SDL_RENDERER_PRESENTVSYNC); -- 0)
        if not renderer then
            gulp_cpp.debugMessage("Vsync renderer not avaiable. SDL says:\n")
            gulp_cpp.debugMessage(SDL_GetError())
            gulp_cpp.debugMessage("\n")
            -- accept any
            renderer = gulp_cpp.SDL_CreateRenderer(GetGameWindow(), renderer_index, 0)
            vsync_guess = false
        end
    end
    if not renderer then
        gulp_cpp.debugMessage("Could not create renderer. SDL says:\n")
        gulp_cpp.debugMessage(SDL_GetError())
        gulp_cpp.debugMessage("\n")
        return 1
    end
    
    -- clear the screen immediately so that any start up delay is 'mitigated'
    -- on Linux windows construction, so that we don't end up with horrid 'background' window.
    gulp_cpp.SDL_RenderClear(renderer)
    gulp_cpp.SDL_RenderPresent(renderer)


    app:SetRenderer(renderer, vsync_guess)

    -- return nil
end
