-- Forlorn Fox main file
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
gulp.LuaStateCodeMode = "CLIENT"

-- uncomment to get a small amount of debug information about level loading
-- gulp.DebugLevelLoading = true

-- engine version check
local required_engine = 0.56
if gulp.forlorn_fox_engine_version < required_engine then
    print("*******************************************")
    print("Engine too old - needs " .. required_engine .. ", currently "..gulp.forlorn_fox_engine_version)
    print("*******************************************")
    os.exit(1)
end

local check_debugger = require("debug_select")
check_debugger("ui")

if argk['--OPEN_STD_CONSOLE'] then -- and gulp.platform ~= "iPhone" then
    -- allow commands from stdin
    gulp.lua_main:open_console()
end

local hotswap_reloader = require("utilities/hotswap_reloader")
if argk['--HOTSWAP_RELOADER'] then
    hotswap_reloader.enable() 
end

-- Check if keepalive disable command is passed in
gulp.enableKeepAlive = true
if argk['--NOKEEPALIVE'] then
    -- turn off the keep alive
    gulp.enableKeepAlive = false
end


-- require statements
require("sound_man")

utf8 = require("utilities/gulp_utf8")
require("helper")

local GlyphEditor = require("glyph_editor/glyph_editor")
local FunctionKeyMonitor = require("utilities/function_key_monitor")


local glyph_edit = GlyphEditor:new()

require("graphics/graphics")

require("mobs/mob_list")


require('map_control')
ClientMapLoader = require('map_controllers/ClientMapLoader')

gulp.map_loader = ClientMapLoader('central_cavern')

-- Networking
GulpNetworking = require("gulp_networking")
gulp_networking = GulpNetworking:new()
-- End Networking

main_menu = require("main_menu/main_menu")
require("networking/netstat")
local file_utils = require("utilities/file_utils")
require('utils')

local action_input = require("action_input")
local MouseMove = require("player/mouse_move")

local function decode_command_line_arguments()

    -- arg[0] = program, arg[1] = first parameter, arg[2] = second parameter, etc.
    --
    -- NOTE: We don't use #arg because (according to Lua 5.2 reference "the length 
    -- of a table t is only defined if the table is a sequence, that is, the set of 
    -- its positive numeric keys is equal to {1..n} for some integer n."
    -- "Note, however, that non-numeric keys do not interfere with whether a table 
    -- is a sequence."

    -- CLIENT mode must supply IP and PORT
    -- SERVER mode must supply PORT
    -- SIZE must specify a size (as a glyph size multiplier - 1 is very small, 5 is pretty big)


    -- set the gulp variables with default values
    gulp.ServerIP = "localhost"
    gulp.start_map_name = 'treehouse'
    gulp.command_line_autorun = nil
    gulp.NetworkError = false
    gulp.gameName = "My World"
    
    if arg[1] == nil then
        print("No command line args passed in, using defaults")
    else
        local client_index = argk['CLIENT']
        local server_index = argk['SERVER']
        if client_index then
            gulp.command_line_autorun = 'CLIENT'
            gulp.target_IPAddr = arg[client_index+1]
            gulp.target_port = arg[client_index+2]
        elseif server_index then
            gulp.command_line_autorun = 'SERVER'
            gulp.prefs.network_port = arg[server_index+1]
            
        end
        
        local size_index = argk['SIZE']
        if size_index then
            gulp_cpp.set_screen_size(arg[size_index+1])
        end 
        
        local map_index = argk['MAP']
        if map_index then
            gulp.start_map_name = arg[map_index+1]
        end
        
        if not argk['--autoplay'] then
            gulp.command_line_autorun = nil
        end
    end
end

local prefs = require("utilities/prefs")
local dispatcher = require("utilities/dispatcher")
require("gulp_constants")
local LocalConnectionFakeSocket = require("networking/local_connection_fake_socket")

function gulp.set_wrap_limits()
    local gr = gulp.app:get_standard_graphics_context()
    -- Minus 0.9 to ensure text can be seen at edge of screen
    gr:set_wrap_limits(gulp_cpp.current_line_start, gulp_cpp.current_line_end, 
        gulp_cpp.current_column_start, gulp_cpp.current_column_end-0.9)
end

local LuaMainThreaded = require("utilities/LuaMainThreaded")

-- Pointless to override/delegate, because it's only called before everything :-)
function gulp.load(app)
    
    
    ---[[
    local num_r = gulp_cpp.RendererInfo.number_renderers
    local current_renderer = gulp_cpp.RendererInfo(gulp.app:get_SDL_renderer(), 0)
    --print(string.format("Current: %s (%s)", current_renderer.name, render_flag_decode(current_renderer)) )
    for i = 0, num_r-1 do
        local rend = gulp_cpp.RendererInfo(nil, i)
        local current = ""
        if current_renderer.name == rend.name then
            current = "ACTIVE"
        end
        print(string.format("Renderer %i: %s (%s) %s", i, rend.name, render_flag_decode(rend), current))
    end
    --]]

    gulp.time = 0
        
    prefs.load()
    
    gulp.app = app

    load_graphics()
    load_sounds()
    
    action_input.init_movement_keys()
    MouseMove.init_game_regions(app)
    
    gulp._delegate_list = dispatcher()
    
    gulp._user_movement_target_list = {}
    gulp.game_managers = {}
    
    file_utils.create_save_folders()
    
    set_absolute_draw_list(app:get_absolute_draw_list())
    
    decode_command_line_arguments()
    
    app:get_standard_graphics_context():set_game_screen_offset(true)
    
    gulp.function_key_monitor = FunctionKeyMonitor:new()

    gulp.main_menu = main_menu()
    
    -------------------------------------------------------------------------------
    gulp.ui_queue = gulp_cpp.LuaStateQueue("ui_queue")
    gulp.net_queue = gulp_cpp.LuaStateQueue("net_queue")
    gulp.server_queue = gulp_cpp.LuaStateQueue("server_queue")
    
    -------------------------------------------------------------------------------
    -- Set up other states as globals
    gulp.lua_game_world = LuaMainThreaded("scripts/game_world_server/server_main.lua")
    
        -- called once on game load
    gulp.lua_game_world:run("server_load", gulp.ui_queue, gulp.net_queue, gulp.server_queue)
    
    gulp.lua_game_world:run_in_thread("server_thread")

   -------------------------------------------------------------------------------
    --gulp.lua_network_low_level = gulp_cpp.LuaMain()    
    --app:setup_ff_lua_state(gulp.lua_network_low_level)
    
    --[[gulp_cpp.load_main_file(gulp.lua_network_low_level, "scripts/networking/network_main.lua")
    
    gulp_cpp.LuaMain_push_queue(gulp.lua_network_low_level, gulp.ui_queue)
    gulp_cpp.LuaMain_push_queue(gulp.lua_network_low_level, gulp.net_queue)
    gulp_cpp.LuaMain_push_queue(gulp.lua_network_low_level, gulp.server_queue)
    gulp_cpp.run_gulp_function_if_exists(gulp.lua_network_low_level, "net_load", 3, 0)   -- called once on game load
    --]]
    
    --gulp.net_thread = LuaThread(gulp.lua_network_low_level, "netthread", "netthread");
    --print("Net thread return value", gulp.net_thread:WaitThread())
    -------------------------------------------------------------------------------

    gulp.local_connection_fake_socket = LocalConnectionFakeSocket:new(gulp.server_queue)

    --[[ -- Queue test with no return Qs
    print("test 1")
    local q = gulp_cpp.LuaStateQueue("main_queue")
    if not q:is_empty() then print("Error", q:is_empty()) end
    q:send(true)
    if q:is_empty() then print("Error", q:is_empty()) end
    local t = q:read()
    print(q, type(t), t)
    if(t ~= true) then print("Error") end
    
    print("test 2")
    -- try sending to the newq
    local newq = gulp_cpp.LuaStateQueue("secondary_queue")
    if not newq:is_empty() then print("Error", newq:is_empty()) end
    newq:send("Hello")
    if newq:is_empty() then print("Error", newq:is_empty()) end
    local t = newq:read()
    print(q, newq, type(t), t)
    if(t ~= "Hello") then print("Error") end
    
    print("test 3")
    -- try sending to the newq
    if not newq:is_empty() then print("Error", newq:is_empty()) end
    q:send("Goodbye")
    if newq:is_empty() then print("Error", newq:is_empty()) end
    local t = q:read()
    print(q, newq, type(t), t)
    print(q:get_identifier(), newq:get_identifier())
    if(t ~= "Goodbye") then print("Error") end
    
    print("test 4")
    if not q:is_empty() then print("Error", q:is_empty()) end
    q:send(1234)
    if q:is_empty() then print("Error", q:is_empty()) end
    local t = q:read()
    print(type(t), t)
    if(t ~= 1234) then print("Error") end

    print("test 5")
    if not q:is_empty() then print("Error", q:is_empty()) end
    q:send(nil)
    if q:is_empty() then print("Error", q:is_empty()) end
    local t = q:read()
    print(type(t), t)
    if(t ~= nil) then print("Error") end

    print("test 6")
    if not q:is_empty() then print("Error", q:is_empty()) end
    q:send({ 1, 2, 3, { fred = 39, sid = 45, "hi", 
                [10]="five", ["!!__!!"] = "O_o", ["gee"]="GEE",
                [""]="Hello",
                { [true] = 999, [false] = 123 }, { true, false } } } )
    if q:is_empty() then print("Error", q:is_empty()) end
    local t = q:read()
    print(type(t), t)
    print_table(t)
    --]]
    --[[ -- Queue test with return Qs
    local q = gulp_cpp.LuaStateQueue("main_queue")
    print(q:is_empty())
    q:send(q, true)
    print(q:is_empty())
    local newq, t = q:read()
    print(q, newq, type(t), t)
    if(t ~= true) then print("Error") end
    
    -- try sending to the newq
    print(newq:is_empty())
    q:send(newq, "Hello")
    print(newq:is_empty())
    local newq2, t = newq:read()
    print(q, newq, newq2, type(t), t)
    if(t ~= "Hello") then print("Error") end
    
    -- try sending to the newq
    print(newq:is_empty())
    q:send(newq, "Goodbye")
    print(newq:is_empty())
    local newq3, t = q:read()
    print(q, newq, newq2, newq3, type(t), t)
    print(q:get_identifier(), newq:get_identifier(), newq2:get_identifier(), newq3:get_identifier())
    if(t ~= "Goodbye") then print("Error") end
    
    print(q:is_empty())
    q:send(q, 1234)
    print(q:is_empty())
    local newq, t = q:read()
    print(type(t), t)
    if(t ~= 1234) then print("Error") end

    print(q:is_empty())
    q:send(q, nil)
    print(q:is_empty())
    local newq, t = q:read()
    print(type(t), t)
    if(t ~= nil) then print("Error") end

    print(q:is_empty())
    q:send(q, { 1, 2, 3, { fred = 39, sid = 45, "hi", 
                [10]="five", ["!!__!!"] = "O_o", ["gee"]="GEE",
                [""]="Hello",
                { [true] = 999, [false] = 123 }, { true, false } } } )
    print(q:is_empty())
    local newq, t = q:read()
    print(type(t), t)
    print_table(t)
    --]]
    
    gulp.set_wrap_limits()
        
end



--local _angle = 0

local function handle_local_network_packet_ui(message)
    if message.mt then
        local dispatcher = gulp.client_receivers[message.mt]
        if dispatcher then
            dispatcher(message)
        end
    else
        print("UI: No mt supplied in network message?")
    end
end

local showingError = false
local function handle_internet_broadcast_error(message)
    -- If not already showing the menu then show it
    if showingError == false then
        gulp.error_menu = Menu {
            background = gulp.colour.BRIGHT_RED.simple_colour, 
            foreground = gulp.colour.BRIGHT_WHITE.simple_colour,
            border_style = 2,
            { message.error_string },
            { "" },
            { "Restart game to try again" },
            { "" },
    --                        { " QUIT to Menu ", function() showingError = false gulp.save_and_quit_game() end },
            { " Continue", function() 
                    showingError = false 
                    end },
        }
        gulp.error_menu.on_close = function()
            showingError = false
        end
    end
    
    -- Stop us showing the network error again
    showingError = true
end

local function handle_internet_broadcast_now_registered()
  -- If the error message is being shown close
  if showingError then
    gulp.error_menu:close()
    showingError = false
  end  
end

local function handle_server_couldnt_start(message)
    
    --gulp.temp_menu = 
    local m = Menu {
            line = 1.5,
            column = 1.5,
            background = gulp.colour.BRIGHT_RED.simple_colour, 
            foreground = gulp.colour.BRIGHT_WHITE.simple_colour,
            border_style = 2,
            { "Couldn't start server" },
            { "" },
            { "Port already in use!" }, --"Failed to bind to port" },
            { "Is another game running?" },
            { "" },
            { " OK", 1 }
        }
    m.on_close = function()
        gulp.temp_menu = nil  
        
        -- we know the local server IS NOT running
        gulp.local_server_started = false

        gulp.end_game()                    
        
        end

        
end

--
-- Queue command dispatch table
-- 
local ui_function_dispatch = {
    qlocal = handle_local_network_packet_ui,
    internet_broadcast_error = handle_internet_broadcast_error,
    internet_broadcast_now_registered = handle_internet_broadcast_now_registered,
    server_couldnt_start = handle_server_couldnt_start,
}

local particle_effects = require("graphics/particle_effects")

local server_dt_update = { qcommand = "update", dt = 0 }

function gulp.update(dt)
    --[[
    gulp.debug:set_lua_info_string(string.format("%f %f %f   %f %f %f", 
            gulp_cpp.current_column_start, 
            gulp_cpp.current_columns, 
            gulp_cpp.current_column_end, 
            gulp_cpp.current_line_start, 
            gulp_cpp.current_line_end,
            gulp_cpp.current_lines))
    --]]
    
    --if gulp.localplayer then
    --    gulp.debug:set_lua_info_string(string.format("L = %f, C = %f", gulp.localplayer.line, gulp.localplayer.column))
    --    gulp.debug:set_lua_info_string(string.format("clo = %f, cco = %f", gulp.localplayer.current_line_offset, gulp.localplayer.current_column_offset))
    --end
    
    -- tell the server as soon as possible, so that the latency is minimised
    server_dt_update.dt = dt
    gulp.server_queue:send(server_dt_update)
        
    gulp.time = gulp.time + dt

    while not gulp.ui_queue:is_empty() do
        local message = gulp.ui_queue:read()
        --print("UI:", message)
        local f = ui_function_dispatch[message.qcommand]
        if f then
            f(message)
        else
            error("unknown command in gulp.update() " .. tostring(f))
        end
    end
    --[[
    _angle = _angle + (36 * dt) 
    if _angle > 360 then
        _angle = _angle - 360
    end
    --]]
    
    --gulp_cpp.LuaMain_push_number(gulp.lua_game_world, dt);
    --gulp_cpp.run_gulp_function_if_exists(gulp.lua_game_world, "server_update", 1, 0);
    --gulp_cpp.LuaMain_push_number(gulp.lua_network_low_level, dt  );
    --gulp_cpp.run_gulp_function_if_exists(gulp.lua_network_low_level, "net_update", 1, 0);
    
    -- Kenny added temp step tick code for copas
    -- make sure receive stuff is done before objects are updated
    gulp_networking:stepCopas()
    
    gulp._delegate_list:update_dispatcher(dt)

    -- make sure send stuff is sent straight away
    --gulp_networking:stepCopas()
    
    hotswap_reloader.monitor_loop(dt)
    
    particle_effects.update(dt)
end


local function blank_borders(gr, colour, maze_line_offset)
    
    local maze_start = maze_line_offset
    local maze_end = maze_start + gulp_cpp.GetAvailableLevelHeight() 
    local left_x = gulp_cpp.current_column_start
    local right_x = gulp_cpp.GetAvailableLevelWidth() + 1
    local screen_start = gulp_cpp.current_line_start
    local screen_end = gulp_cpp.current_line_end
    
    -- c, x1, y1, x2, y2
    gr:FillRect(colour, left_x, screen_start, right_x, maze_start)
    gr:FillRect(colour, left_x, maze_end, right_x, screen_end)
end

local border_colour = get_SDL_Color_by_name('BLACK')

function gulp.draw(gr)
    --[[
    gr:printExT(2 ,2 ,6 ,33, { angle = _angle} )
    gr:printExT(1 ,4 ,5 ,66, { flip = 1 } )
    gr:printExT(1 ,6 ,4 ,66, { flip = 2 } )
    gr:printExT(1 ,8 ,3 ,66, { flip = 3 } )
    gr:printExT(1 ,10 ,6 ,33, { angle = _angle, scale_x = 0.5 + (_angle/36), scale_y = 0.5 + (_angle/36) } )
    gr:printExT(1 ,12 ,3 ,65, { angle = _angle, scale_x = 4 } )
    --]]
    set_graphics_context(gr)
    
    local maze_line_offset = gulp.TOP_UI_HEIGHT

    -- draw the map and the status lines - if enabled
    if gulp._draw_map and gulp.localplayer then
        local map_name = gulp.localplayer.map_name
        if check_map_exists(map_name) then
            local map_to_render = get_maze(map_name)
            if map_to_render then
              --check_mdls(map_to_render, "draw")
              map_to_render:print(gr, maze_line_offset)
              particle_effects.draw(gr)
              blank_borders(gr, border_colour, maze_line_offset)
            end
        end
    end
    
    if gulp._draw_status_lines then
        gulp.app:print_status_lines(gr, maze_line_offset);
    end
    
    local ret = gulp._delegate_list:passive_dispatcher('draw', gr)
    
    MouseMove.draw_regions(gr)

    -- draw UI elements etc
    render_absolute_draw_list(gr, 0)

    
    render_debug_data(gr)
    return ret
end

function gulp.mousepressed(x, y, handled, right)
    return gulp._delegate_list:action_dispatcher('mousepressed', x, y, handled, right)
end

function gulp.mousereleased(x, y, handled, right)
    return gulp._delegate_list:action_dispatcher('mousereleased', x, y, handled, right)
end

function gulp.mouse_down_moved(x, y, handled, right)
    return gulp._delegate_list:action_dispatcher('mouse_down_moved', x, y, handled, right)
end

function gulp.keypressed(sym, scancode, modifier, repeated)
    local handled = gulp._delegate_list:action_dispatcher('keypressed', sym, scancode, modifier, repeated)
    
    if not handled then
        action_input.game_key_processing(scancode, true)
    end
end

function gulp.keyreleased(sym, scancode, modifier)
    local handled = gulp._delegate_list:action_dispatcher('keyreleased', sym, scancode, modifier)
    
    if not handled then
        action_input.game_key_processing(scancode, false)
    end
end


function gulp.textinput(...)
    return gulp._delegate_list:action_dispatcher('textinput', ...)
end

function gulp.textediting(...)
    return gulp._delegate_list:action_dispatcher('textediting', ...)
end



-- does this make sense to delegate?
function gulp.focus(...) -- f)
    return gulp._delegate_list:action_dispatcher('focus', ...)
end

function gulp.launch_glyph_editor()
    glyph_edit:launch_glyph_editor()
end


Player = require("player/Player")
require("main_menu/connect_menu")


function gulp._start_game_after_connect()
    MouseMove.enable_game_controls()
    gulp._draw_map = true
    gulp._draw_status_lines = true
    
    local player_id = nil
    
    -- client has a slave map loader
    gulp.map_loader:LoadInitialMap()
    
    -- create the player
    -- must be after the initial maze is loaded (nasty dependency!!)
    -- because the player accesses the C++ maze data directly
    -- in various places, including its initialize() function
    gulp.localplayer = Player(gulp.selected_avatar, gulp.selected_colour, player_id)
    gulp.add_user_movement_target(gulp.localplayer)
    
    -- Running in client mode
    GulpNetworking:send_register_player_from_Client(gulp.localplayer)
    
    -- must be done after player registered
    gulp.map_loader:request_neighbouring_maps(gulp.localplayer.map_name)
    
    -- delegate for networking - this will change to be 
    -- "run networking in separate lua context" and will always happen (no check)
    -- after refactoring
    -- TODO
    gulp._delegate_list:add_delegate(gulp_networking)
    
end

function gulp.close_all_movetargets()
    for _,v in pairs(gulp._user_movement_target_list) do
        -- if close exists in that movement target
        if v.close then
            v:close()
        end
    end
end

function gulp.save_and_quit_game()
    gulp.save_game()
    gulp_cpp.SetLevelSize(gulp.MINIMUM_COLUMNS_ON_SCREEN, gulp.MINIMUM_LINES_ON_SCREEN - (gulp.TOP_UI_HEIGHT+gulp.BOTTOM_UI_HEIGHT))
    gulp.end_game()
    remember_game_end_for_debug()   
end

function gulp.save_game()
end


-- This might be called on a partially constructed game ...
-- so we tear down what ever has been constructed. It is also safe to call twice, 
-- e.g. from a local server message after the client has already aborted itself.
function gulp.end_game(reason)
    
    particle_effects.delete_all()

    if gulp.main_menu then
        gulp.main_menu:close()          -- close if not closed already
    end
    
    MouseMove.disable_game_controls()
    gulp._draw_map = false
    gulp._draw_status_lines = false

    -- close down server at some appropriate point 
    if gulp.local_server_started then
        gulp.local_server_started = false
        gulp.server_queue:send( { qcommand = "stop" } )
    end
    
    -- drop all the players of all types
    for _,player in all_players() do
        player:remove()
    end
    
    
    gulp_networking:unregisterClient()
    
    gulp.close_all_movetargets()
    gulp.remove_user_movement_target(gulp.localplayer)
    
    -- drop the maps
    gulp._delegate_list:drop_delegate(gulp.map_loader)
    gulp.map_loader:DropAllMaps()
    
    if gulp.localplayer then
        gulp.localplayer:remove()
        gulp.localplayer = nil
    end
    gulp.movetarget = nil
    
    -- make sure we remove all the dropped graphical elements
    -- from the screen before we (potentially) restart the game
    collectgarbage()
    
    gulp.main_menu = main_menu()
    if reason then
        gulp.reason_for_quit_menu = Menu{
                line = 4,
                column = 4,
                background = gulp.colour.BLUE.simple_colour, 
                foreground = gulp.colour.BRIGHT_WHITE.simple_colour,
                border_style = 2,
                { reason },
                { "" },
                { "  OK  ", 99 }
            }
    end
    
    -- End network stuff
    gulp_networking:disableClient()
    
    gulp._delegate_list:drop_delegate(gulp_networking)
  
    -- clean up to avoid reuse later
    gulp.target_IPAddr = nil
    gulp.target_port = nil
    
    if gulp.text_buffer then
        gulp.text_buffer:clear()
    end
    
    -- reset back to registration mode so we can connect to a new server
    client_receiver_set_register_mode()
end


function gulp.add_user_movement_target(object)
    
    -- send a 'key up' to the current movetarget before we switch to the new one
    if gulp.movetarget ~= nil and gulp.movetarget.action then
        gulp.movetarget:action(0)
    end
    
    
    table.insert(gulp._user_movement_target_list, object)
    gulp.movetarget = gulp._user_movement_target_list[#gulp._user_movement_target_list]
end

function gulp.remove_user_movement_target(object)
    -- removes the last copy only
    local delegates = gulp._user_movement_target_list
    local i = #gulp._user_movement_target_list
    while i > 0 do
        if delegates[i] == object then
            table.remove(delegates, i)
            break
        end
        i = i - 1
    end
    
    -- make sure we have a movetarget that's in the list
    local len = #gulp._user_movement_target_list
    if len then
        gulp.movetarget = gulp._user_movement_target_list[len]
    else
        gulp.movetarget = nil
    end
end

function gulp.user_movement_not_handled(object, movement_type, ...)
    -- if object is in the list twice, this is a problem...
    
    print("haven't implemented this yet")
    os.exit(1) -- not implemnted yet
end

function gulp.quit_event()
    -- check if not save world, if not, save
end

function gulp.quit()
    -- tell threads to quit
    gulp.server_queue:send( { qcommand = "stop_thread" } )
    
    -- wait up to 0.5 for the server thread to respond
    for _ = 1,10 do
        gulp_cpp.SDL_Delay(50)
        if not gulp.lua_game_world:IsRunning() then
            break
        end
    end
    if gulp.lua_game_world:IsRunning() then
        error("Thread ".. gulp.lua_game_world:GetThreadName() .. "Still running on shutdown")
    end
    
    --print("Server Game World thread return value", gulp.lua_game_world:WaitThread())
    -- Not strictly necessary, but we do it anyway...
    gulp.lua_game_world:WaitThread()
end


function gulp.unknown_event(type)
    --print("Unknown event", type)
end

--
-- SDL_WindowEventID values
--
local SDL_WINDOWEVENT_NONE = 0          -- Never used
local SDL_WINDOWEVENT_SHOWN = 1         -- Window has been shown
local SDL_WINDOWEVENT_HIDDEN = 2        -- Window has been hidden
local SDL_WINDOWEVENT_EXPOSED = 3       -- Window has been exposed and should be redrawn
local SDL_WINDOWEVENT_MOVED = 4         -- Window has been moved to data1, data2
local SDL_WINDOWEVENT_RESIZED = 5       -- Window has been resized to data1xdata2
local SDL_WINDOWEVENT_SIZE_CHANGED = 6  -- The window size has changed, either as a result of an API call or through the system or user changing the window size
local SDL_WINDOWEVENT_MINIMIZED = 7     -- Window has been minimized
local SDL_WINDOWEVENT_MAXIMIZED = 8     -- Window has been maximized
local SDL_WINDOWEVENT_RESTORED = 9      -- Window has been restored to normal size and position
local SDL_WINDOWEVENT_ENTER = 10        -- Window has gained mouse focus
local SDL_WINDOWEVENT_LEAVE = 11        -- Window has lost mouse focus
local SDL_WINDOWEVENT_FOCUS_GAINED = 12 -- Window has gained keyboard focus
local SDL_WINDOWEVENT_FOCUS_LOST = 13   -- Window has lost keyboard focus
local SDL_WINDOWEVENT_CLOSE = 14        -- The window manager requests that the window be closed

function gulp.windowevent(event, data1, data2)
    --print("Window event", event, data1, data2)
    if event == SDL_WINDOWEVENT_SIZE_CHANGED then
        --print("Resized to", data1, "x", data2)
        gulp_cpp.UpdateScreenData_game_window()
        
        if gulp.localplayer then
            gulp.localplayer:calculate_offsets()
        end
        
        local enabled = MouseMove.disable_game_controls()
        MouseMove.init_game_regions(gulp.app)
        MouseMove.restore_game_controls(enabled)
    end
    
    -- Do this when any window event happens.
    gulp.set_wrap_limits()
    
    --[[
    SDL_WINDOWEVENT_RESIZED
    window has been resized to data1xdata2; this is event is always preceded by 
    SDL_WINDOWEVENT_SIZE_CHANGED

    SDL_WINDOWEVENT_SIZE_CHANGED
    window size has changed, either as a result of an API call or through the system 
    or user changing the window size; this event is followed by 
    SDL_WINDOWEVENT_RESIZED if the size was changed by an external event, i.e. the 
    user or the window manager
    --]]
end



function gulp.setNetworkError()
    gulp.NetworkError = true 
end    

function gulp.clearNetworkError()
    gulp.NetworkError = false 
end    

function gulp.isNetworkError()
    return gulp.NetworkError
end    

function server_run_lua(command)
    gulp.server_queue:send( { qcommand = "run_lua", lua = command } )
end
