-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--

require("strict")
local class = require("middleclass")
local TextEffects = require("utilities/text_effects")

local FallingLetters = class("FallingLetters")

function FallingLetters:set_letter(row, column, letter)
    local t = TextEffects:new(row, column, letter)
    local size = 200 + math.random(0, 100)
    t:small(size)
    t:colour("BRIGHT_MAGENTA")
    local angle = -10 + math.random(0,20)
    t:rotate(angle)
    t.FL_size_inc = (math.random(0,1) * 2)-1
    t.FL_size_inc = t.FL_size_inc * 10
    t.FL_angle_inc = (math.random(0,1) * 2)-1
    t.FL_angle_inc = t.FL_angle_inc * 5
    t.FL_size = size
    t.FL_angle = angle
    return t
end

function FallingLetters:set_fall()
    local colours = {
        gulp.colour.RED.simple_colour,
        gulp.colour.GREEN.simple_colour,
        gulp.colour.YELLOW.simple_colour,
        gulp.colour.BLUE.simple_colour,
    }
    
    local t = { 
        row = math.random(math.floor(gulp_cpp.current_line_start),math.floor(gulp_cpp.current_line_end)),
        column = math.random(math.floor(gulp_cpp.current_column_start), math.floor(gulp_cpp.current_column_end)),
        --char = math.random(33, 127),
        char = 0x17, -- evil egg glyph
        colour = random_entry(colours),
        speed = math.random(1,11)
    }
    return t
end

function FallingLetters:initialize()
    gulp._delegate_list:add_delegate(self)

    self.fall = {}
    for i = 0, 20 do
        table.insert(self.fall, self:set_fall())
    end
    
    local letters = {}
    --[[
    table.insert(letters, self:set_letter(18, 3, "F"))
    table.insert(letters, self:set_letter(18, 5.2, "o"))
    table.insert(letters, self:set_letter(18, 7.4, "r"))
    table.insert(letters, self:set_letter(18, 9.6, "l"))
    table.insert(letters, self:set_letter(18, 11.8, "o"))
    table.insert(letters, self:set_letter(18, 14, "r"))
    table.insert(letters, self:set_letter(18, 16.2, "n"))
    table.insert(letters, self:set_letter(18, 22.4, "F"))
    table.insert(letters, self:set_letter(18, 24.6, "o"))
    table.insert(letters, self:set_letter(18, 26.8, "x"))
    --]]
    table.insert(letters, self:set_letter(5, 7.4, "E"))
    table.insert(letters, self:set_letter(6, 9.6, "g"))
    table.insert(letters, self:set_letter(5, 11.8, "g"))
    table.insert(letters, self:set_letter(5, 18, "T"))
    table.insert(letters, self:set_letter(6, 20.2, "a"))
    table.insert(letters, self:set_letter(5, 22.4, "g"))
    
    self._letters = letters
end

function FallingLetters:draw(gr)
--    print_string_at(20, 0, tostring(self._letters[1].FL_angle))
--    print_string_at(21, 0, tostring(self._letters[1].FL_angle_inc))
--    set_fg_colour("BLUE")
    set_bg_colour_transparent()
    for _, f in ipairs(self.fall) do
        set_fg_colour(f.colour)
        print_glyph_at(f.row, f.column, f.char)
    end
    
    --[[   Remove Whale/Dragon/Test drawing ... looks weird without it :-)
    set_fg_colour("MAGENTA")
    print_glyph_at(10,10,graphic_shorthand("Dragon.png",33))
    
    print_glyph_at(10,11,graphic_shorthand("Dragon.png",128))
    print_glyph_at(10,12,graphic_shorthand("Dragon.png",129))
    print_glyph_at(10,13,graphic_shorthand("Dragon.png",130))
    print_glyph_at(10,14,graphic_shorthand("Dragon.png",131))
    print_glyph_at(10,15,graphic_shorthand("Dragon.png",132))
    print_glyph_at(10,16,graphic_shorthand("Dragon.png",133))
    print_glyph_at(10,17,graphic_shorthand("Dragon.png",134))
    print_glyph_at(10,18,graphic_shorthand("Dragon.png",135))
    
    print_glyph_at(11,11,graphic_shorthand("Dragon.png",144))
    print_glyph_at(11,12,graphic_shorthand("Dragon.png",145))
    print_glyph_at(11,13,graphic_shorthand("Dragon.png",146))
    print_glyph_at(11,14,graphic_shorthand("Dragon.png",147))
    print_glyph_at(11,15,graphic_shorthand("Dragon.png",148))
    print_glyph_at(11,16,graphic_shorthand("Dragon.png",149))
    print_glyph_at(11,17,graphic_shorthand("Dragon.png",150))
    print_glyph_at(11,18,graphic_shorthand("Dragon.png",151))
    --]]
    
end

function FallingLetters:update(dt)
    for _, letter in ipairs(self._letters) do
        
        local size_inc = dt * letter.FL_size_inc 
        letter.FL_size = letter.FL_size + size_inc
        if letter.FL_size > 300 then
            letter.FL_size_inc = -math.abs(letter.FL_size_inc)
        elseif letter.FL_size < 200 then
            letter.FL_size_inc = math.abs(letter.FL_size_inc)
        end
        letter:small(letter.FL_size)
        
        local angle_inc = dt * letter.FL_angle_inc
        if letter.FL_angle > 10 then
            letter.FL_angle_inc = -math.abs(letter.FL_angle_inc)
        elseif letter.FL_angle < -10 then
            letter.FL_angle_inc = math.abs(letter.FL_angle_inc)
        end
        letter.FL_angle = letter.FL_angle + angle_inc
        letter:rotate(letter.FL_angle)
    end
    
    for i, f in ipairs(self.fall) do
        f.row = f.row + dt * f.speed
        if f.row > gulp_cpp.current_line_end then
            self.fall[i] = self:set_fall()
            self.fall[i].row = gulp_cpp.current_line_start-2
        end
    end
    
end

function FallingLetters:close()
    for _, letter in ipairs(self._letters) do
        letter:hide()
    end
    gulp._delegate_list:drop_delegate(self)
end

return FallingLetters
