-- Forlorn Fox
-- Copyright © 2014 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
require("strict")
local class = require("middleclass")
local copas = require("copas")

-- Needed to create the remote player
local PlayerUpdatingAvatar = require("player/PlayerUpdatingAvatar")
local Mob = require("controllers/Mob")
local chat = require("player/chat")

-- required for new_network_object
local PickableManager = require('pickable_manager')

local SpeechBubble = require('objects/speech_bubble')

local clientSocket = nil        -- no socket to start with


function GulpNetworking:sendPlayerPosition(col, line, map)  
    local location_delta_required = gulp.minimum_cell_movement_before_transmit_in_cells
    local send = false

    -- If the player has moved
    --if (col ~= self.lastSentColumn) or (line ~= self.lastSentLine) then
    if (math.abs(line - self.lastSentLine) > location_delta_required) or
        (math.abs(col - self.lastSentColumn) > location_delta_required) or
        (self.last_tagged ~= gulp.localplayer.tagged) then
    
        -- store the new postion
        self.lastSentColumn = col
        self.lastSentLine = line
        self.last_tagged = gulp.localplayer.tagged
        self.moving = true
        
        send = true
    
    elseif self.moving then
        -- we were moving but now we're not
        -- send one last location message to turn off the animation on other clients
        self.moving = false
        
        send = true
    
    end
    
    if send then
        -- If networking client mode enabled
        if (clientSocket and self.registered) then
            local positionMessage = gulp.localplayer:form_position_message()
            clientSocket:SendMessage(positionMessage);
        end
        
    end     
end 

function GulpNetworking:UpdateLastSentPosition(player)
    self.lastSentLine = player.line
    self.lastSentColumn = player.column
end

function GulpNetworking:SendToServer(message)
    clientSocket:SendMessage(message);
end
    
function GulpNetworking:SendDroppedItem(line, column, item)

    if (clientSocket and self.registered) then

        local message = { 
        mt = "dropped",
        line = line,
        column = column,
        item = item:serialise()         
        }
    
        clientSocket:SendMessage(message);
        
        if gulpNetworkingDebuggingEnabled then
            print("SendDroppedItem", message)
        end
    end
end

function GulpNetworking:SendTaggedPlayers(tagged_player_id, untagged_player_id)

    if (clientSocket and self.registered) then
        
        local message = { 
            mt = "player_tagged",
            tagged_player = tagged_player_id,
            untagged_player = untagged_player_id,
        }
    
        clientSocket:SendMessage(message);
    end
        
end

function GulpNetworking:SendSpeech(player_id, text_in)

    if (clientSocket and self.registered) then
        
        local message = { 
            mt = "player_speech",
            player = player_id,
            text = text_in,
        }
    
        clientSocket:SendMessage(message);
    end
        
end



function GulpNetworking:sendPlayerAction(action)
    -- Check if the connection was made
    if (clientSocket and self.registered) then     
    
        local actionMessage = { 
            mt = "player_action",
            action = action
        }

        clientSocket:SendMessage(actionMessage);
       
        -- Debug messages
        if gulpNetworkingDebuggingEnabled then
            print("sendPlayerAction", actionMessage)
        end  
    end   
end 

function GulpNetworking:send_request_connect_to_server()
    if (clientSocket) then     
        local message = {
            mt = "request_connnect",
            magic = "FORLORN FOX CONNECT",
            version = gulp.forlorn_fox_version,
        }
        clientSocket:SendMessage(message)
        return true
    end
    return false
end

function GulpNetworking:send_register_player_from_Client(player)                    

    -- Check if the connection was made
    if (clientSocket) then     
        
        local Message = {
            mt = "register_player",
            name = gulp.prefs.player_name,
            avatar = player.avatar,
            colour = player.colour,
            line = player.line,
            column = player.column,
            map_name = player.map_name
        }
        
        clientSocket:SendMessage(Message)
    end
    
end     



function GulpNetworking:unregisterClient()                  
    
    -- Check if the connection was made
    if (clientSocket and self.registered) then      
      
      --local Message = ("<unregister>" .. gulp.localplayer.id .. "</unregister>\r\n")

        local Message = {
            mt = "unregister",
            id = gulp.localplayer.id
            }

        -- Send the message
        clientSocket:SendMessage(Message)

        -- Debug messages
        if gulpNetworkingDebuggingEnabled then
            print("unRegisterClient", Message)
        end  
    end
    return true
end


function GulpNetworking:SendFunctionKey(key)
    local message = {
        mt = "function_key",
        key = key
    }
    -- Send the message
    clientSocket:SendMessage(message)

    -- Debug messages
    if gulpNetworkingDebuggingEnabled then
        print("SendFunctionKey", message)
    end  
end



-- make a class that has it's own variables
ConnectorStore = class("ConnectorStore")

function ConnectorStore:initialize()
    self._abort_connect = false
end

function ConnectorStore:stop_connect()
    self._abort_connect = true
end

function GulpNetworking:enableClient(ServerIPAddress, ServerPort)   
    
    if clientSocket then
        gulp.connecting_menu:tell_result(false, "Already connected")
    end
    
    if self.target_server_to_connect_to_IP_address or 
        self.target_server_to_connect_to_port or clientSocket then
            -- probably want to register a callback function, but for the moment, just 
            -- callback directly
            -- gulp.connecting_menu:tell_result(false, "Already trying to connect")
            
        self.target_server_to_connect_to_IP_address = nil
        self.target_server_to_connect_to_port = nil
        self.last_connector:stop_connect()
    end
    
    -- Ensure we have the register message handlers loaded
    client_receiver_set_register_mode()
    
    self.target_server_to_connect_to_IP_address = ServerIPAddress
    self.target_server_to_connect_to_port = ServerPort
    
    self.last_connector = ConnectorStore:new()
    
    local mythread = copas.addthread(ClientReceiveThread) -- Listens for messages
    -- add name in ClientRecieveThread itself
end


function GulpNetworking:set_local_connect_socket()
    if clientSocket then
        error("client Socket already exists in set_local_connect_socket()?")
    end
    clientSocket = gulp.local_connection_fake_socket
end

function GulpNetworking:finalise_client_connect()
    local success = false
    
    -- Now register the client 
    if self:send_request_connect_to_server() then  
        --if gulpNetworkingDebuggingEnabled then 
        --    print("GulpNetworking:enableClient IP:", ServerIPAddress, "Port:", ServerPort)
        --end 
       
        success = true
    else
        self:disableClient()
        success = false
        error_string = "Failed to send connect request because no socket"
    end
    
    return success
end

local socket = require("socket")
local NetworkTransport = require("networking/network_transport")


function ConnectorStore:non_blocking_connect(networking)   
        
    local ServerIPAddress = networking.target_server_to_connect_to_IP_address
    local ServerPort = networking.target_server_to_connect_to_port
    
    local master_socket = socket.tcp()
    if not master_socket then
        gulp.connecting_menu:tell_result(false, "Couldn't make new socket")
        return false
    end
    local mythread = coroutine.running()
    name_network_item(mythread, "thread - ClientReceiveThread for ".. tostring(master_socket))
    
    -- Create copas socket for transmitting and receive
    local success, err = copas.connect(master_socket, ServerIPAddress, ServerPort)
    if self._abort_connect then
        print("Aborted connect (in non_blocking_connect())")
        return false
    end
    print("status in non_blocking_connect() from Copas.connect", success, err)
    
    -- check a clientSocket doesn't exist already...
    if clientSocket then
        master_socket:close()       -- ensure closed
        gulp.connecting_menu:tell_result(false, "client Socket is already open??")
        return false
    end
    
    local internal_clientSocket = nil
    if success then
        internal_clientSocket = master_socket
    end
    
    remember_socket_for_debug("client", internal_clientSocket)
    
    print("enableClient running on socket: ", internal_clientSocket)
    
    local success = false
    local error_string = "No error recorded"
    if (internal_clientSocket) then 
 
        -- option not exposed via copas wrapper, so go direct
        local setoption_success = master_socket:setoption('tcp-nodelay', true)

        if not setoption_success then
            print("Couldn't set TCP-Nodelay")
        end

        networking.ServerIP = ServerIPAddress
        networking.Port = ServerPort
         
        internal_clientSocket:settimeout(0)
        clientSocket = NetworkTransport(internal_clientSocket)
        
        local mythread = copas.addthread(GulpNetworking.SendMessageThread)   -- Used to send messages
        name_network_item(mythread, "addthread - client SendMessageThread ")

        success = networking:finalise_client_connect()

    else
        gulp_networking:disableClient()
        success = false
        error_string = "No Server?"
    end

    -- probably want to register a callback function, but for the moment, just 
    -- callback directly
    gulp.connecting_menu:tell_result(success, error_string)
    
    networking.target_server_to_connect_to_IP_address = nil
    networking.target_server_to_connect_to_port = nil
        
    return success
end



function GulpNetworking:disableClient()
    if self.last_connector then
        self.last_connector:stop_connect()
    end
    if (clientSocket) then
        GulpNetworking.AddSocketToClose(clientSocket)
        clientSocket = nil
    end
end


local function client_process_mob_message(message)
    local id = message.id
    local map_name = message.map_name
    local mob = gulp.maps[map_name].auto.mobs[id]
            
    if mob == nil then
        gulp.maps[map_name].auto.mobs[id] = NewMob(message)
    else
        mob:set_location(message.line, message.column)
    end

end


-- forward declaration
client_receiver_set_online_mode = nil

local function client_register_return(message)
    
    if gulpNetworkingDebuggingEnabled then
        print("RegisterClient Message response: ")
        print_table(message)
        print("RegisterClient response status:", status)  
    end

    gulp_networking.registered = true
    gulp.localplayer:set_id(message.player_id)
    gulp.localplayer:move_to_start_location()
    
    if (gulp.localplayer.id) then
      if gulpNetworkingDebuggingEnabled then
        print("Player Registered: ", gulp.localplayer.id)   
      end 
    else
      print("Error RegisterClient - invalid response")
      -- @todo: disconnect somehow
    end
end

local function client_registration_complete(message)
    -- add things here that can't be handled in
    -- client_register_return
    
    -- the difference is, at this point, we have the full list of other players
    gulp.localplayer:change_tag_game_mode(true)
end



local function client_reject_connect_return(message)
    if not message.reason then
        message.reason = "?"
    end
    if gulpNetworkingDebuggingEnabled then
        print("rejected connect:", reason)
    end
    gulp_networking:disableClient()
    gulp.client_register_failed(message.reason, message.server_version)
end

local function client_accept_connect_return(message)
    
    if message.magic and message.magic == "Forlorn Fox Connect" then
        if gulpNetworkingDebuggingEnabled then
            print("accepted connect")
        end
        client_receiver_set_online_mode()
        gulp.client_register_succeeded()
        
        -- Start checking for the server going missing with the keep alive
        gulp_networking.keepAlive:resetServerPingTimeout()
        gulp_networking.keepAlive:startCheckingForMissingServer()
    else
        if gulpNetworkingDebuggingEnabled then
            print("accepted connect - BAD MAGIC")
        end
        gulp_networking:disableClient()
        gulp.client_register_failed("magic")  
    end
end
        
local function client_process_game_control_message(message)
    find_game_manager_by_name(message.manager):process_game_control_message(message)
end

local function client_process_player_control_message(message)
    gulp.localplayer:process_player_control_message(message)
end

local function client_process_player_move_message(message)
    local id = message.id
                        
    if gulpNetworkingDebuggingEnabled then 
        print("Move message received: ", player_id)
    end  

    -- If the message is complete then pass on to process
    if id then
        --print(id)
        local player = get_player_by_id(id)
        --print(player)
        if player then
            player:processPosition(message)
        end
    end 
end    
     
local function client_handle_player_tagged_message(message)
    local tagged_id = message.tagged_player
    local untagged_id = message.untagged_player
    
    if tagged_id then
        local player = get_player_by_id(tagged_id)
        if player then
            player:process_tagged('tag')
        end
    end 
    
    if untagged_id then
        local player = get_player_by_id(untagged_id)
        if player then
            player:process_tagged('untag')
        end
    end 
end

local function client_process_player_item_picked_message(message)
    gulp.localplayer:process_pick_item_message(message.item)
end


local function client_process_player_die_message(message)
    -- Close the socket the Copas thread will catch this and tidyup
    gulp.localplayer:die()
end    
  
      
local function client_process_server_exit_message(message)
    -- Close the socket the Copas thread will catch this and tidyup
    -- don't need this .. gulp.end_game does it anyway.
    --clientSocket:close()

    -- Stop looking for the missing server
    gulp_networking.keepAlive:stopCheckingForMissingServer()
    
    gulp.end_game("Server Quit") 
end        
 
 
local function client_process_mob_visibility_message(message)
    local id = message.id
    local map_name = message.map_name
    
    -- The server sometimes sends messages for mobs in maps we don't 
    -- have loaded, during the process of loading and dropping maps.
    -- Make sure we don't crash because of it.
    if gulp.maps[map_name] == nil then
        return
    end
    
    local mob = gulp.maps[map_name].auto.mobs[id]
            
    if mob ~= nil then
        mob:process_visibility_message(message)
    end
end    
 
local function client_process_mob_location_message(message)
    local id = message.id
    local map_name = message.map_name
    
    -- The server sometimes sends messages for mobs in maps we don't 
    -- have loaded, during the process of loading and dropping maps.
    -- Make sure we don't crash because of it.
    if gulp.maps[map_name] == nil then
        return
    end
    
    local mob = gulp.maps[map_name].auto.mobs[id]
            
    if mob ~= nil then
        mob:process_location_message(message)
    end
end    
       
local function client_process_mob_glyph_message(message)
    local id = message.id
    local map_name = message.map_name
  
    -- The server sometimes sends messages for mobs in maps we don't 
    -- have loaded, during the process of loading and dropping maps.
    -- Make sure we don't crash because of it.
    if gulp.maps[map_name] == nil then
        return
    end
    
    local mob = gulp.maps[map_name].auto.mobs[id]
            
    if mob ~= nil then
        mob:process_glyph_message(message)
    end
end    
 
 
local function client_process_floor_item_message(message)
    local map_name = message.map_name
    local id = message.network_id
    if gulp.maps[map_name].auto.pickables[id] ~= nil then
        print("I've been sent a floor_item that already exists:", map_name, id)
    else
        gulp.maps[map_name].auto.pickables[id] = new_network_object(message)
    end
end

local function client_process_item_show_message(message)
    local map_name = message.map_name
    local id = message.network_id
    
    local pickable = gulp.maps[map_name].auto.pickables[id]
    if pickable ~= nil then
        pickable:process_show_message(message)
    end
end

local function client_process_item_hide_message(message)
    local map_name = message.map_name
    local id = message.network_id
    local pickable = gulp.maps[map_name].auto.pickables[id]
    if pickable ~= nil then
        pickable:process_hide_message(message)
    end
end
   
local function client_update_conditional(message)
    local map = get_map(message.map_name)
    map.conditionals[message.conditional] = message.value
end

local function client_glyph_update(message)
    local map = get_map(message.map_name)
    local maze = map.auto.maze
    local line = message.line
    local column = message.column
    local glyph = message.glyph
    
    map[line][column] = glyph
    maze:set_item(line-1, column-1, glyph)
end

   
local function client_process_add_remote_ghost_message(message)

    -- Create the remote player for this client, if it does not already
    -- exist. It might already exist if this a broadcast from the server
    -- this client's own player.
    if not does_player_exist(message.id) then
        PlayerUpdatingAvatar(message.avatar, message.colour, message.id, message.map_name, message.line, message.column, message.tagged,
                        message.tool, message.tool_orientation)

        if gulpNetworkingDebuggingEnabled then 
            print("Ghost message registered: ", message.id)
                
            -- Print out the list of registed clients
            for key,value in all_remote_players() do print(key,value.avatar) end
        end  
    end
    
end  
   
local function client_process_remove_remote_ghost_message(message)

    -- Create the remote player for this client
    local player = get_player_by_id(message.id)
    if player then
        player:remove()
    end
    
    --if gulpNetworkingDebuggingEnabled then 
        print("Ghost message remove: ", message.id)
            
        -- Print out the list of registed clients
        for key,value in all_remote_players() do print(key,value.avatar) end
    --end  
end  
   
   
local function client_process_play_sound_message(message)
    play_located_sound(message.sound, message.volume, message.line, message.column, message.map)
end

local function client_process_chat_message(message)
    if gulpNetworkingDebuggingEnabled then
        print("client_process_chat_message: ", message.message)
    end
    chat.print_chat_message(message.message)
end


local function client_process_server_ping_message(message)
    local response = { 
        mt = 'server_ping_ack',
    }
    
    clientSocket:SendMessage(response);
  
    -- Reset the timeout for as we have seen the server ping us
    gulp_networking.keepAlive:resetServerPingTimeout()
    
    if gulpNetworkingDebuggingEnabled then
        print("client_process_server_ping_message")
    end
end

local function client_process_map_file(message)
    if gulpNetworkingDebuggingEnabled then
        print("client_process_map_file("..message.map_file_name..")")
    end
    
    gulp.map_loader:load_map_file(message.map_file_name, message.file)
    
    -- load all the maze assets
    -- these might need to be updated even before we enter the level (e.g. slime growth)
    load_maze_assets(message.map_file_name)
    
end

local function client_process_initial_map_file(message)
    if gulpNetworkingDebuggingEnabled then
        print("client_process_initial_map_file("..message.map_file_name..")")
    end
    
    gulp.client_initial_map_name = message.map_file_name
    gulp.client_initial_map_file = message.file 
end

local function client_process_reset_map(message)
    print("client resetting map "..message.map_name)
    
    local map_name = message.map_name
    local loader = gulp.map_loader
    
    -- drop the map and immediately request it from the server again
    loader:DropMap(map_name)
    clientSocket:SendMessage(loader:form_map_request_message(map_name))
end


local function client_process_debug_print(message)
    if not gulp.server_debug_print_table then
        gulp.server_debug_print_table = {}
    end
    for k,v in pairs(message.data) do
        gulp.server_debug_print_table[k] = v
    end
end

local function client_process_player_speech(message)
    local player_id = message.player
    local text = message.text
            
    local player = get_player_by_id(player_id)
    if player then
        player:process_speech(text)
    end
end

-- this table allow specific dispatch of message types (mt)
local _client_online_receivers = 
{
    accept_register = client_register_return,
    registration_complete = client_registration_complete,
    game_control = client_process_game_control_message,
    player_move = client_process_player_move_message,
    player_control = client_process_player_control_message,
    server_exit = client_process_server_exit_message,
    die = client_process_player_die_message,
    picked = client_process_player_item_picked_message,
    mob = client_process_mob_message,
    mob_visibility = client_process_mob_visibility_message,
    mob_location = client_process_mob_location_message,
    mob_glyph = client_process_mob_glyph_message,
    floor_item = client_process_floor_item_message,
    item_show = client_process_item_show_message,
    item_hide = client_process_item_hide_message,
    add_remote_ghost_id = client_process_add_remote_ghost_message,
    remove_remote_ghost_id = client_process_remove_remote_ghost_message,
    play_sound = client_process_play_sound_message,
    player_tagged = client_handle_player_tagged_message,
    chat_message = client_process_chat_message,  
    server_ping = client_process_server_ping_message,  
    map_file = client_process_map_file,
    reset_map = client_process_reset_map,
    conditional_update = client_update_conditional,
    glyph_update = client_glyph_update,
    debug_print = client_process_debug_print,
    player_speech = client_process_player_speech,
    
}

local _client_register_receivers = 
{
    reject_connect = client_reject_connect_return,
    accept_connect = client_accept_connect_return,
    map_file = client_process_initial_map_file,
}

-- actual table to dispatch from
gulp.client_receivers = _client_register_receivers


function client_receiver_set_register_mode()
    gulp.client_receivers = _client_register_receivers
end

function client_receiver_set_online_mode()
    gulp.client_receivers = _client_online_receivers
end


-- Receive thread for the client data. Listens for messages then may or may not send response.
function ClientReceiveThread()
    local success = gulp_networking.last_connector:non_blocking_connect(gulp_networking)
    if not success then
        print("non_blocking_connect failed, so returning from ClientReceiveThread()")
        return
    end
    
    local message = decode_from_network_incremental_new()
   
    while true do 
        -- Check there is a socket is not just return
        -- this is what happens at game end 
        if (clientSocket == nil) then 
            return
        else
            -- Now wait on the response to check the client registed ok
            local reqdata, err, partial = copas.receive(clientSocket.my_socket, "*l")  -- read just a line
                -- only get partial back after an error
                
            -- Check the connection didn't have an error
            if (err) then
                if (err == "closed") then 
                print ("ClientReceiveThread Copas socket closed so exiting") 
            
                else
                    print ("ClientReceiveThread Copas receive error stopping err was: ", err) 
                end  
                
                GulpNetworking.HandleNetworkErrorClient(clientSocket, "ClientReceive")
                
                -- Set the socket to nil as we know it is closed
                clientSocket = nil
            return
        end
    
        -- Check for partial message recevied
        -- Will probably never get here because err will be set.
        --if (partial) then
        --  print ("ClientReceiveThread Copas receive parial") 
        --end
        
        -- I don't think we need this if-reqdata test, because if it's nil
        -- then we must have an error message
        --if reqdata then
            local finished = decode_from_network_incremental_add_line(message, reqdata)
            if finished and message.mt then
                local dispatcher = gulp.client_receivers[message.mt]
                if dispatcher then
                    if message.map_name and not message.always_process_at_client and gulp.maps[message.map_name] == nil then
                        --print("Client got message for map_name: " .. message.map_name .. " but does not have that map loaded")
                    else
                        dispatcher(message)
                    end
                end
                message = decode_from_network_incremental_new()
            end     
        --end
  
        end
    end
end  

