-- Forlorn Fox
-- Copyright © 2017 Rob Probin
-- No redistribution without permission.
--
-- This acts like a protected LuaMain
--

require("strict")
local class = require("middleclass")

local LuaMainThreaded = class("LuaMainThreaded")

function LuaMainThreaded:initialize(main_file)
    self._internal_lua_state = gulp_cpp.LuaMain()
    
    -- load the librares, C-Lua bridge stuff, etc.
    gulp.app:setup_ff_lua_state(self._internal_lua_state)
    
    gulp_cpp.load_main_file(self._internal_lua_state, main_file)
end

function LuaMainThreaded:run(function_to_run, ...)
    local num_params = 0
    for k,v in ipairs{...} do
        gulp_cpp.LuaMain_push_queue(self._internal_lua_state, v)
        num_params = num_params + 1
    end
    gulp_cpp.run_gulp_function_if_exists(self._internal_lua_state, function_to_run, num_params, 0)
end

function LuaMainThreaded:run_in_thread(function_to_run)
    self._our_thread = gulp_cpp.LuaThread(self._internal_lua_state, function_to_run, function_to_run)
    --print("Server thread return value", self._our_thread:WaitThread())
end

function LuaMainThreaded:IsRunning()
    return self._our_thread:IsRunning()
end

function LuaMainThreaded:GetThreadName()
    return self._our_thread:GetThreadName()
end

function LuaMainThreaded:WaitThread()
    local thread = self._our_thread
    self._our_thread = nil          -- ensure we don't reference it again
    return thread:WaitThread()
end

function LuaMainThreaded:GetThreadID()
    return self._our_thread:GetThreadID()
end




return LuaMainThreaded

