-- Forlorn Fox
-- Copyright © 2015 Rob Probin, Tony Park and Kenny Osborne.
-- No redistribution without permission.
--
-- Prefs
--
-- Created by Rob Probin, 1st September 2015
-- For Forlorn Fox

local lfs = require"lfs"

local prefs = {}

local prefs_defaults =
{
    network_port = 10000,       -- DEFAULT_NETWORK_PORT
--    player_name = function_to_generate_unique_player_name
    player_name = "Player" .. tostring(math.random(100000)),
    inhibit_broadcast = false,          -- broadcast to LAN is generally ok
    allow_internet_broadcast = false,       -- don't broadcast to the world by default
}

local prefs_filename = "prefs.dat"
function prefs.save()
    local f, err = io.open(gulp_cpp.SaveDataPath(prefs_filename):str(), "w")
    if not f then
        print(err)
        return
    end
    f:write("local data={\n")
    for k,v in pairs(gulp.prefs) do
        if type(k) == "string" then
            local value_str
            if type(v) == "string" then
                value_str = string.format("%q", v)
            elseif type(v) == "number" then
                value_str = string.format("%g", v)
            elseif type(v) == "boolean" then
                value_str = tostring(v)                
            else
                break
            end
            local line
            if k:match("[%a_]+[%a%d_]*") then
                line = string.format(" %s = %s,\n", k, value_str)
            else
                line = string.format(" [%q] = %s,\n", k, value_str)
            end
            f:write(line)
        end
    end
    f:write("} return data\n")
    f:close()
end

function prefs.load()
    local filename = gulp_cpp.LoadPath(prefs_filename):str()
    local mode, err = file_utils.attributes(filename, "mode")
    if mode ~= "file" then
        --if mode == nil then
        --    print("Get attributes failed", err)
        --end
        return
    end
    
    local success, data = pcall(dofile, filename)
    if not success then
        print("Error Loading prefs", data)
        return
    end
    
    if type(data) == "table" then
        gulp.prefs = data
        -- fill in missing values
        for k,v in pairs(prefs_defaults) do
            if gulp.prefs[k] == nil then
                gulp.prefs[k] = v
            end
        end
    else
        print("Unexpected data from load prefs", type(data))
    end
end

function prefs.init()
    gulp.prefs = {}
    for k, v in pairs(prefs_defaults) do
        gulp.prefs[k] =v
    end
end

-- always initialise if it doesn't exist
if gulp['prefs'] == nil then
    prefs.init()
end


return prefs
